// gamecharacter.cpp
#include "gamecharacter.h"
#include "constants.h"
#include "manager.h"

#include <alleggl.h>
#include <assert.h>


GameCharacter::GameCharacter(Manager *m) : SpriteObject(m), canTravelOnWater(false)
{
	spriteSet = ResourceManager::SpriteSetEnemies;
	spriteNum = 0;
	spriteScale = PlayerSpriteScale;
	// spriteOffset = - FairySpriteFloorLevel;
	// spriteOffset = - PlayerSpriteFloorLevel ;
	animTimer = 0;
	animFrame = 0;
	// animLength = 2; // fairy anim length
	// firstSpriteNum = 41; // start of fairy anim (S)
	firstSpriteNum = 0; // start of knight anim.
	animLength = 4; // knight /player anim length.
	spriteOffset = - KnightSpriteFloorLevel;
	azimuth = 0;
	desiredAzimuth = 0;
	currentSpeed = 0;
	desiredSpeed = 0;
	maxSpeed = 2.0; // units / sec
	killed = false;
	radius = 0.3;
	hitPoints = 100.0;
}

// Angle 0 = +y
// Angle 90 = +x

void GameCharacter::setSprite()
{
	// azimuth has a range -180 .. 180 so we need to be within that.
	int direction = (int) ((azimuth / 45.0) + 4.5);
	// Because North is actually what the sprites call "NorthWest", we need to
	// subtract one.
	direction = (direction + 7) % 8;
	animTimer = AnimTicks;
	animFrame = (animFrame + 1) % animLength;
	spriteNum = firstSpriteNum + (direction * animLength) + animFrame;
	if (killed) {
		if (animFrame == (animLength -1)) {
			// Stop animation once we reach the end.
			firstSpriteNum += animFrame;
			animFrame = 0;
			animLength = 1;
		}
	}
}

void GameCharacter::runLogic()
{
	if (animTimer > 0) {
		animTimer--;
	} else {
		setSprite();
	}
	if (! killed) {
		runController();
		// If we're not facing the way we want to be, turn so we are.
		if (fabs(desiredAzimuth - azimuth) > AzimuthTolerance ) {
			Component dir = desiredAzimuth - azimuth;
			normaliseAngle(dir);
			if (dir >0) {
				azimuth += TurnRate;
			} else {
				azimuth -= TurnRate;
			}
		} else {
			// Jump to precisely the right place, we're close already.
			azimuth = desiredAzimuth;
		}
		normaliseAngle(azimuth);
		// If we're going too slowly or too fast, adjust.
		if (fabs(desiredSpeed - currentSpeed) > SpeedTolerance) {
			if (desiredSpeed > currentSpeed) {
				currentSpeed += Acceleration;
				if (currentSpeed > maxSpeed)
					currentSpeed = maxSpeed;
			} else {
				currentSpeed -= Acceleration;
				if (currentSpeed < SpeedTolerance)
					currentSpeed = 0;
			}
		}
		// Move in the direction we're going.
		// Remember speed is in units / 100 ticks.
		Vect3d moveVector = Vect3d::fromPolar(azimuth, (currentSpeed / 100));
		location = location + moveVector;
		clip();
		locationChanged();
	}
}

/*
 * virtual method. Overrideen for AI.
 */
void GameCharacter::runController()
{
	// Spin on the spot.
	// and run around!
	/*
	if (key[KEY_W]) {
		desiredSpeed = maxSpeed;
	} else {
		desiredSpeed = 0;
	}
	if (key[KEY_A]) {
		desiredAzimuth = azimuth - 10;
	}
	if (key[KEY_D]) {
		desiredAzimuth = azimuth + 10;
	}
	
	if (key[KEY_K]) {
		setKilled();
	}
	*/
}

void GameCharacter::setKilled()
{
	if (killed) {
		return;
	}
	// Face south - west in our system (South in the sprites)
	azimuth = -135; 
	desiredAzimuth = azimuth;
	// Stop
	currentSpeed = 0;
	killed = true;
	// Go to the death frame.
	firstSpriteNum = deathSpriteNum; // death frame
	animLength = deathAnimLength;
	// Reset animation
	animFrame = 0;
	spriteNum = firstSpriteNum;
	animTimer = AnimTicks;
	collidable = false;
}

// This routine taken from a previous work of mine
// This routine determines if the character is overlapping with a tile, and if so,
// Moves it away enough to no longer overlap.
void GameCharacter::clip()
{
	// work out which block we are in...
	int blockX = ((int) location.x);
	int blockY = ((int) location.y);
	// h and v
	if (isBlocked(blockX, blockY - 1, layer)) {
		clipPoint(Vect3d(location.x, blockY), blockX, blockY - 1);
	}
	if (isBlocked(blockX, blockY + 1, layer)) {
		clipPoint(Vect3d(location.x, blockY + 1),blockX, blockY + 1);
	}
	if (isBlocked(blockX - 1, blockY, layer)) {
		clipPoint(Vect3d(blockX, location.y),blockX - 1, blockY);
	}
	if (isBlocked(blockX + 1, blockY, layer)) {
		clipPoint(Vect3d(blockX + 1, location.y), blockX + 1, blockY);
	}
	// diagonals
	if (isBlocked(blockX - 1, blockY - 1, layer)) {
		clipPoint(Vect3d(blockX, blockY ), blockX - 1, blockY - 1);
	}
	if (isBlocked(blockX + 1, blockY - 1, layer)) {
		clipPoint(Vect3d(blockX + 1, blockY ), blockX + 1, blockY - 1);
	}
	if (isBlocked(blockX - 1, blockY + 1, layer)) {
		clipPoint(Vect3d(blockX ,blockY + 1), blockX - 1, blockY + 1);
	}
	if (isBlocked(blockX + 1, blockY + 1, layer)) {
		clipPoint(Vect3d(blockX + 1 , blockY + 1), blockX + 1, blockY + 1);
	}
}

void GameCharacter::clipPoint(const Vect3d &p, int blockx, int blocky)
{
	Vect3d rel = p - location;
	// If it doesn't in fact overlap, we're ok.
	if (rel.magnitudeSquared() > (radius * radius))
	{
		return;
	}
	// Otherwise we need to do something.
	// work out by HOW MUCH it's overlapping.
	Component dist = rel.magnitude();
	Component overlap = radius - dist; // amount we overlap.
	Vect3d unitVector = rel * (1.0 / dist);
	Vect3d offset = unitVector * (overlap);
	// Now move the character (out of the wall)
	location = location - offset;
	// Register a wall hit.
	hitWall(blockx, blocky);
}

GameCharacter::~GameCharacter()
{

}

