// Map rendering, loading etc routines.

#include "map.h"
#include "manager.h"
#include "enemies.h"
#include <alleggl.h>
#include <math.h>
#include <fstream>

void Map::loadEmptyMap()
{
	// Bottom layer:
	// All outside tiles in the map should be solid.
	// All others should be non-solid.
	// Other layers:
	// All outside tiles should be solid,
	// Other tiles should have no ground (fall-through)
	// Everything should be invisible
	
	for (int z=0; z<MapLayers; z++) {
		for (int y=0; y<MapSize; y++)
		for (int x=0; x<MapSize; x++) {
			bool outside = (x <2 ) || (y < 2);
			outside = outside || (y == (MapSize-1));
			outside = outside || (x == (MapSize-1));
			MapBlock & block = blockAt(x,y,z);
			block.groundTile = 0;
			block.wallTile = 2;
			block.visible = false;
			block.special = NoSpecial;
			if (outside) {
				block.wallType = WallTypeSolid;
			} else {
				block.wallType = WallTypeNone;
				if (z == 0) {
					block.floorType = FloorTypeFlat;
				} else {
					block.floorType = FloorTypeMissing;
				}
			}
			block.groundTile = (x % 4) / 2;
		}
	}
	// Make sure the 0,0,0 is invisible (area outside map).
	blockAt(0,0,0).visible = false;
	
	/* Some test map.
	// A pond.
	for (int y=15; y<25; y++)
	for (int x=5; x<15; x++) {
		MapBlock & block = blockAt(x,y,0);
		block.floorType = FloorTypeWater;
	}
	
	// To keep some interest, here is a test map.
	for (int i=5; i<14; i++) {
		MapBlock & block = blockAt(i,8,0);
		block.wallType = WallTypeSolid;
		MapBlock & block2 = blockAt(8,i,0);
		block2.wallType = WallTypeSolid;
		MapBlock & block3 = blockAt(i,11,0);
		block3.wallType = WallTypeSolid;
	}
	*/
}

void Map::load(Manager *manager, const char * filename)
{
	std::ifstream in(filename);
	if (! in.good()) {
		throw "Unable to open map file or something";
	}
	// Reset the map.
	loadEmptyMap();
	// Read the number of layers.
	int numLayers;
	in >> numLayers;
	for (int layer=0; layer < numLayers; layer++) {
		int rows;
		in >> rows; // Number of rows
		// read floor textures...
		for (int y =(rows - 1); y >=0; y--) {
			char buf[100];
			in.getline(buf,100);
			int len = std::strlen(buf);
			for (int x=0; x< len; x++) {
				char c = buf[x];
				MapBlock & block = blockAt(x,y,layer);
				if (c == 'w') {
					block.floorType = FloorTypeWater;
				} else {
					block.floorType = FloorTypeFlat;
					block.groundTile = c - '0';
				}
			}
		}
		// Now read the map data.
		for (int y =(rows - 1); y >=0; y--) {
			char buf[100];
			in.getline(buf,100);
			int len = std::strlen(buf);
			for (int x=0; x< len; x++) {
				char c = buf[x];
				loadBlock(manager, c, x, y, layer);
			}
		}
	}
}

void Map::loadBlock(Manager *manager, char c, int x, int y, int z)
{
	MapBlock & block = blockAt(x,y,z);
	// 0..3 - solid tile with wall texture 0..3

	if ((c >= '0') && (c <= '9')) {
		block.wallTile = c - '0';
		block.wallType = WallTypeSolid;
		return;
	}
	Vect3d blockCentre = Vect3d(x, y, 0) + Vect3d(0.5, 0.5);
	// Otherwise, we probably don't want a wall here.
	block.wallType = WallTypeNone;
	GameObject * obj = 0;
	switch (c) {
		case 'p':
			// TODO p - player
			manager->relocatePlayer(blockCentre, z);
			break;
		case 'f': // fairy 
			obj = new EnemyFairy(manager);
			break;
		case 'k': // knight 
			obj = new EnemyKnight(manager);
			break;
		case 's':
			 // skull 
			obj = new StaticObject(manager, StaticObject::Skull);
			break;
		case 'S': // skull pile 
			obj = new StaticObject(manager, StaticObject::SkullPile);
			break;
		case 'g': // goat
			obj = new StaticObject(manager, StaticObject::Goat);
			break;
		case 'd':
			block.wallType = WallTypeDoorX;
			break;
		case 'D':
			block.wallType = WallTypeDoorY;
			break;
		case 'a':
			// Decoration item 1 (vase)
			obj = new StaticObject(manager, StaticObject::Vase);
			break;
		case 'e': // exit level
			block.special = ExitSpecial;
			break;
		default:
			break;
	}
	if (obj) {
		obj->setLocation(blockCentre);
		obj->setLayer(z);
		manager->addObject(obj);
	}
}

bool Map::renderBlock(int x, int y, int z)
{
	MapBlock & block = blockAt(x,y,z);
	MapBlock & ne = blockAt(x+1,y,z);
	MapBlock & nw = blockAt(x,y+1,z);
		
	float xf = x;
	float yf = y;
	// Even if block is not visible, draw ne and nw walls if they are.
	if (block.wallType != WallTypeSolid) {
		// If the adjacent blocks to the NE and NW are solid, render the back
		// walls of the block.
		if ((ne.wallType == WallTypeSolid) && ne.visible) {
			// Render ne (+x) wall.
			// Set rendering params.
			glColor3f(0.8, 0.8, 0.8);
			glEnable(GL_TEXTURE_2D);
			// Determine texutre.
			manager->rm.selectTexture(ne.wallTile);
			
			glBegin(GL_TRIANGLE_STRIP);
			glTexCoord2d(0,WallHeight);
			glVertex3f(xf+1, yf, WallHeight);
			glTexCoord2d(1,WallHeight);
			glVertex3f(xf+1, yf +1, WallHeight);
			glTexCoord2d(0,0);
			glVertex3f(xf+1, yf, 0);
			glTexCoord2d(1,0);
			glVertex3f(xf+1, yf+1, 0);
			glEnd();
		}
		if ((nw.wallType == WallTypeSolid) && nw.visible) {
			// Render nw (+y) wall.
			// Set rendering params.
			glColor3f(0.6, 0.6, 0.6);
			glEnable(GL_TEXTURE_2D);
			// Determine texutre.
			manager->rm.selectTexture(nw.wallTile);
			glBegin(GL_TRIANGLE_STRIP);
			glTexCoord2d(0,WallHeight);
			glVertex3f(xf, yf +1, WallHeight);
			glTexCoord2d(0,0);
			glVertex3f(xf, yf +1 , 0);
			glTexCoord2d(1,WallHeight);
			glVertex3f(xf+1, yf+1, WallHeight);
			glTexCoord2d(1,0);
			glVertex3f(xf+1, yf+1 , 0);
			glEnd();
		}
	}
	// STOP HERE if the block is not visible
	if (! block.visible) return false;
	
	if (block.wallType != WallTypeSolid) {
		// Render floor.
		switch (block.floorType) {
			case FloorTypeFlat:
				renderFlatFloor(x,y,z,block);
				break;
				// Slopes: TODO
			case FloorTypeWater:
				renderWaterFloor(x,y,z,block, ne, nw);
				break;
			default:
				// Nothing
				break;
		}
	} 
	const float DoorOffset = 0.5;
	switch (block.wallType)
	{
		case WallTypeSolid:
			// It IS solid hey?
			// Render a "ceiling piece" on.
			// TODO: different shaped ceiling piece depending on block type
			// (triangular blocks).
			glColor3f(0.2,0.2,0.2);
			glDisable(GL_TEXTURE_2D);
			glBegin(GL_TRIANGLE_STRIP);
			glVertex3f(xf, yf +1, WallHeight);
			glVertex3f(xf, yf, WallHeight);
			glVertex3f(xf+1, yf+1, WallHeight);
			glVertex3f(xf+1, yf, WallHeight);
			glEnd();
			break;
		case WallTypeDoorX:
			glColor3f(0.8, 0.8, 0.8);
			// Front of door.
			glDisable(GL_TEXTURE_2D);
			// TODO:Determine texutre.
			// manager->rm.selectTexture(ne.wallTile);
			
			glBegin(GL_TRIANGLE_STRIP);
			glTexCoord2d(0,WallHeight);
			glVertex3f(xf + DoorOffset, yf, WallHeight);
			glTexCoord2d(1,WallHeight);
			glVertex3f(xf + DoorOffset, yf +1, WallHeight);
			glTexCoord2d(0,0);
			glVertex3f(xf + DoorOffset, yf, 0);
			glTexCoord2d(1,0);
			glVertex3f(xf + DoorOffset, yf+1, 0);
			glEnd();
			break;
		
		case WallTypeDoorY:
			// Render nw door
			// Set rendering params.
			glColor3f(0.6, 0.6, 0.6);
			glDisable(GL_TEXTURE_2D);
			// TODO: Determine texutre.
			// manager->rm.selectTexture(nw.wallTile);
			glBegin(GL_TRIANGLE_STRIP);
			glTexCoord2d(0,WallHeight);
			glVertex3f(xf, yf + DoorOffset, WallHeight);
			glTexCoord2d(0,0);
			glVertex3f(xf, yf + DoorOffset , 0);
			glTexCoord2d(1,WallHeight);
			glVertex3f(xf+1, yf + DoorOffset, WallHeight);
			glTexCoord2d(1,0);
			glVertex3f(xf+1, yf + DoorOffset , 0);
			glEnd();
			break;
		default:
			// nothing
			break;
	}
	
	// (todo) compute floor offset
	// (if the floor type != FloorTypeMissing)
	// If the block is a diagonal piece facing the camera, render the diagonal piece.
	return true;
}

void Map::renderFlatFloor(int x, int y, int z, const MapBlock & block)
{
	float xf = x;
	float yf = y;
	// Render the floor.
	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
	manager->rm.selectTexture(block.groundTile);
	// TODO: Set more rendering params.
	glBegin(GL_TRIANGLE_STRIP);
	glTexCoord2d(0,1);
	glVertex3f(xf, yf +1, 0);
	glTexCoord2d(0,0);
	glVertex3f(xf, yf, 0);
	glTexCoord2d(1,1);
	glVertex3f(xf+1, yf+1, 0);
	glTexCoord2d(1,0);
	glVertex3f(xf+1, yf, 0);
	glEnd();
	
}

void Map::renderWaterFloor(int x, int y, int z, 
	const MapBlock & block, const MapBlock & ne, const MapBlock & nw)
{
	float xf = x;
	float yf = y;
	const float WaterHeight = - 0.2;
	const float BottomHeight = -0.4;
	// if ne block or nw block are not water,
	// draw the side of the pool.
	if (ne.floorType != FloorTypeWater) {
		// Render ne (+x) wall.
		// Set rendering params.
		glColor3f(0.8, 0.8, 0.8);
		glEnable(GL_TEXTURE_2D);
		// Determine texutre.
		manager->rm.selectTexture(ne.wallTile);
		
		glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2d(0,0);
		glVertex3f(xf+1, yf, 0);
		glTexCoord2d(1,0);
		glVertex3f(xf+1, yf +1, 0);
		glTexCoord2d(0,BottomHeight);
		glVertex3f(xf+1, yf, BottomHeight);
		glTexCoord2d(1,BottomHeight);
		glVertex3f(xf+1, yf+1, BottomHeight);
		glEnd();
	}
	if (nw.floorType != FloorTypeWater) {
		// Render nw (+y) wall.
		// Set rendering params.
		glColor3f(0.6, 0.6, 0.6);
		glEnable(GL_TEXTURE_2D);
		// Determine texutre.
		manager->rm.selectTexture(nw.wallTile);
		glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2d(0,0);
		glVertex3f(xf, yf +1, 0);
		glTexCoord2d(0,BottomHeight);
		glVertex3f(xf, yf +1 , BottomHeight);
		glTexCoord2d(1,0);
		glVertex3f(xf+1, yf+1, 0);
		glTexCoord2d(1,BottomHeight);
		glVertex3f(xf+1, yf+1 , BottomHeight);
		glEnd();
	}
		
	// Render the floor.
	glColor3f(0.6,0.6,1);
	glDisable(GL_TEXTURE_2D);
	// Enable lighting for the water.
	glEnable(GL_LIGHTING);
	int offset = x*3 + y;
	float height1 = waterHeightMap[offset % 16] * WaterHeight  + WaterHeight;
	Vect3d normal1 = waterNormalMap[offset % 16];
	float height2 = waterHeightMap[(offset + 1) % 16] * WaterHeight  + WaterHeight;
	Vect3d normal2 = waterNormalMap[(offset + 1) % 16];
	float height3 = waterHeightMap[(offset + 3) % 16] * WaterHeight  + WaterHeight;
	Vect3d normal3 = waterNormalMap[(offset + 3) % 16];
	float height4 = waterHeightMap[(offset + 4) % 16] * WaterHeight  + WaterHeight;
	Vect3d normal4 = waterNormalMap[(offset + 4) % 16];
	glBegin(GL_TRIANGLE_STRIP);
	// glBegin(GL_LINE_LOOP); // debug
	glNormal3f(normal2.x, normal2.y, normal2.z);
	glVertex3f(xf, yf +1, height2);
	glNormal3f(normal1.x, normal1.y, normal1.z);
	glVertex3f(xf, yf, height1);
	glNormal3f(normal4.x, normal4.y, normal4.z);
	glVertex3f(xf+1, yf+1, height4);
	glNormal3f(normal3.x, normal3.y, normal3.z);
	glVertex3f(xf+1, yf, height3);
	glEnd();
	// Turn lighting off for everything else.
	glDisable(GL_LIGHTING);
}

Component WaveSpeed = 0.5;

// Sets up a sinusoidal water height map in waterHeightMap
void Map::runLogic()
{
	waveTimer += 1.0/100.0;
	if (waveTimer > 8) {
		waveTimer = 0;
	}
	Component wavePos = WaveSpeed * waveTimer * M_PI * 2;
	for (int i=0; i < 16; i++) {
		waterHeightMap[i] = sin(wavePos);
		wavePos += M_PI / 8;
		waterNormalMap[i] = Vect3d(0,0,1) + (Vect3d(0.6,0.3,0) * cos(wavePos));
	}
}


struct MapPoint {
	int x,y;
	MapPoint(int i, int j) : x(i), y(j) { }
};

typedef std::vector<MapPoint> MapPointList;
typedef MapPointList::iterator MapPointListIterator;

/* 
 * This is a sort of flood fill. It stops when it reaches walls / doors etc.
 *
 * When the player opens a door, the area on all sides of it becomes visible.
 */
void Map::makeTileVisible(int x,int y, int z)
{
	MapPointList points;
	MapPointList newpoints;
	points.push_back(MapPoint(x+1,y));
	points.push_back(MapPoint(x-1,y));
	points.push_back(MapPoint(x,y+1));
	points.push_back(MapPoint(x,y-1));
	while (points.size() > 0) {
		for (MapPointListIterator i = points.begin(); i != points.end(); i++) {
			MapPoint point = *i;
			MapBlock & block = blockAt(point.x, point.y, z);
			if (! block.visible) {
				block.visible = true;
				// If it's passable, do the adjacent ones too.
				if (block.wallType == WallTypeNone) {
					newpoints.push_back(MapPoint(point.x + 1, point.y));
					newpoints.push_back(MapPoint(point.x - 1, point.y));
					newpoints.push_back(MapPoint(point.x, point.y + 1));
					newpoints.push_back(MapPoint(point.x, point.y - 1));
					// Diagonals.
					newpoints.push_back(MapPoint(point.x + 1, point.y +1));
					newpoints.push_back(MapPoint(point.x - 1, point.y -1));
					newpoints.push_back(MapPoint(point.x + 1, point.y - 1));
					newpoints.push_back(MapPoint(point.x - 1, point.y + 1));
				}
			}
		}
		points = newpoints;
		newpoints.clear();
	}

}


