// playercharacter.cpp
#include "playercharacter.h"
#include "constants.h"
#include "manager.h"
#include "shot.h"

#include <alleggl.h>
#include <cstdlib>

static const int PlayerStandingSprite = 0;
static const int PlayerWalkingSprite = 8;
static const int PlayerAttackingSprite = 40;
static const int AttackStartTime = 25;
static const int AttackRecoverTime = 25;
static const int StopTime = 0;
static const Component SmallDistance = 0.4;
static const Component StoppingDistanceFactor = 1.5;
static const Component MinSpeed = 0.25;
static const Component MeleeDamageAmount = 19.0;
static const Component ShooterDamageAmount = 39.0;
static const Component MeleeRadius = 0.1;
static const Component MeleeReach = 0.5;

PlayerCharacter::PlayerCharacter(Manager *m) : GameCharacter(m)
{
	spriteSet = ResourceManager::SpriteSetMainChar;
	spriteScale = PlayerSpriteScale;
	firstSpriteNum = 8; // start of troll anim.
	spriteNum = firstSpriteNum;
	animLength = 4; // knight /player anim length.
	spriteOffset = - PlayerSpriteFloorLevel;
	maxSpeed = 3.0; // units / sec
	currentState = Stopped;
	nextState = Stopped;
	nextStateTime = 0;
	// Defaults seeing as we don't have one.
	deathSpriteNum = 0;
	deathAnimLength = 1;
	previousHitPoints = hitPoints;
	painLevel = 0;
	ammo = 15;
	currentWeapon = Skullshooter;
}

/*
 * virtual method. Overridden by player.
 */
void PlayerCharacter::runController()
{
	Vect3d toTarget = targetLocation - location;
	// Set desiredSpeed.
	Component targetDistance = toTarget.magnitude();
	if (currentState == Walking) {
		// Work out stopping dist.
		desiredSpeed = targetDistance * StoppingDistanceFactor;
		if (desiredSpeed < MinSpeed) {
			desiredSpeed = MinSpeed;
		}
	} else {
		desiredSpeed = 0;
	}
	// Set desired Azimuth if we're walking around.
	if (currentState == Walking) {
		if (targetDistance > SmallDistance) {
			// Target is significantly far away.
			desiredAzimuth = toTarget.angle();
		} else {
			nextState = Stopped;
			nextStateTime = StopTime;
		}
	}
	
	if (nextState != currentState)
	{
		if (nextStateTime > 0) {
			nextStateTime --;
		} else {
			// Switch state.
			currentState = nextState;
			switch (currentState) {
				case Stopped:
					{
						firstSpriteNum = PlayerStandingSprite;
						animLength = 1;
						// Check if we've exited the level.
						MapBlock block = manager->map.blockAt((int) location.x, 
							(int) location.y, layer);
						if (block.special == ExitSpecial) {
							manager->levelComplete = true;
						}
						// Cheat:
						if (key[KEY_F1]) manager->levelComplete = true;;
						break;
					}
				case Walking:
					firstSpriteNum = PlayerWalkingSprite;
					animLength = 4;
					break;
				case AttackRecover:
					doAttack();
					firstSpriteNum = PlayerStandingSprite;
					animLength = 1;
					nextState = Stopped;
					nextStateTime = AttackRecoverTime;
					break;
				default:
					// AttackStart is not actually used here.
					break;
			}
			animFrame = 0;
			animTimer = AnimTicks;
			spriteNum = firstSpriteNum;
			setSprite();
		}
	}
	// sort out pain level (for screen flashing).
	if (hitPoints < previousHitPoints) {
		painLevel += (previousHitPoints - hitPoints);
	} else {
		if (painLevel > 0) {
			painLevel -= 0.1;
		}
	}
	previousHitPoints = hitPoints;
}

void PlayerCharacter::doAttack()
{
	if (ammo == 0) {
		currentWeapon = Fist;
	}
	switch (currentWeapon) {
		case Fist:
		{
			// Cause some damage!
			Component damage = MeleeDamageAmount +
				std::rand() % 6;
			Vect3d meleeTarget(location + Vect3d::fromPolar(azimuth, MeleeReach));
			manager->areaDamage(meleeTarget, damage, 
				MeleeRadius, this);
			break;
		}
		case Skullshooter:
		{
			float ShotSpeed = 4.0;
			Shot *shot = new Shot(manager , location, Vect3d::fromPolar(azimuth, ShotSpeed));
			shot->skull = true;
			shot->setOwner(this);
			shot->setDamage(ShooterDamageAmount + std::rand() % 6);
			manager->addObject(shot);
			ammo --;
		}
	}
}

PlayerCharacter::~PlayerCharacter()
{

}

void PlayerCharacter::setTarget(const Vect3d &target)
{
	if ((currentState != AttackStart) && (currentState != AttackRecover)) {
		targetLocation = target;
		nextState = Walking;
	}
}

void PlayerCharacter::beginAttack(const Vect3d &target)
{
	if ((currentState != AttackStart) && (currentState != AttackRecover)) {
		// Begin the attack.
		currentState = AttackStart;
		nextState = AttackRecover;
		nextStateTime = AttackStartTime;
		// Face our target.
		Vect3d toTarget = target - location;
		desiredAzimuth = toTarget.angle();
		desiredSpeed = 0;
		// Set up sprite.
		firstSpriteNum = PlayerAttackingSprite;
		animLength = 1;
		// Init the actual sprite.
		animFrame = 0;
		animTimer = AnimTicks;
		setSprite();
		
	} else {
		// Naff state for attack.
	}
}


// For opening doors.
void PlayerCharacter::hitWall(int blockx, int blocky) 
{
	MapBlock & block = manager->map.blockAt(blockx, blocky, layer);
	if ((block.wallType == WallTypeDoorX) || (block.wallType == WallTypeDoorY)) {
		block.wallType = WallTypeNone; 
		manager->map.makeTileVisible(blockx, blocky, layer);
	}
}
void PlayerCharacter::handleCollision(const Vect3d &amount, GameObject *what)
{
	GameCharacter::handleCollision(amount, what);
	if (what->isGettable()) {
		// Aha, it's gettable.
		// Get it.
		what->gotBy(this);
	}
}

void PlayerCharacter::setKilled()
{
	// Change it into an enemy body because we don't have our own death seq.
	spriteSet = ResourceManager::SpriteSetEnemies;
	deathSpriteNum = 62;
	deathAnimLength = 1;
	GameCharacter::setKilled();
}

