#include "resourcemanager.h"
#include "gameoptions.h"
#include <math.h>

ResourceManager::~ResourceManager()
{
	if (mainfont) {
		allegro_gl_destroy_font(mainfont);
		mainfont = 0;
	}
	if (fontsfile) {
		unload_datafile(fontsfile);
		fontsfile = 0;
	}
}

void ResourceManager::load()
{
	// Load fonts.
	fontsfile = load_datafile("fonts.dat");
	if (! fontsfile) {
		throw "Unable to load fonts file";
	}
	// Note - scale needs to be small because 1.0 is a big unit in our world.
	FONT * originalfont = (FONT*)fontsfile[0].dat;
 	mainfont = allegro_gl_convert_allegro_font(originalfont,
		AGL_FONT_TYPE_TEXTURED, 64.0);
	if (! mainfont) {
		throw "Unable to convert font";
	}
	loadSprites();
	loadTextures();
}

void ResourceManager::loadTextures()
{
	PALETTE pal;
	BITMAP * bmp = load_bitmap("gfx/textures.tga", pal);
	if (! bmp) {
		throw "Unabel to load textures";
	}
	int numtextures = bmp->w / 64;
	for (int i=0; i<numtextures; i++) {
		BITMAP * temp = create_sub_bitmap(bmp, i* 64,0, 64, 64);
		GLuint texnum = allegro_gl_make_texture_ex( 0, // flags
			temp,
			GL_RGB);
		if (texnum == 0) {
			throw "Unable to convert texture.";
		}
		textures.push_back(texnum);
		destroy_bitmap(temp);
	}
	// Create an extra fake texture for the level end area.
	BITMAP * temp = create_sub_bitmap(bmp, 2* 64,0, 64, 64);
	int colour = makecol(255,255,0);
	circlefill(temp, 32, 32, 12, colour);
	GLuint texnum = allegro_gl_make_texture_ex( 0, // flags
		temp,
		GL_RGB);
	if (texnum == 0) {
		throw "Unable to create special texture";
	}
	textures.push_back(texnum);
	destroy_bitmap(temp);
	
	destroy_bitmap(bmp);
}

void ResourceManager::selectTexture(int n)
{
	GLuint texnum = textures[n];
        glBindTexture(GL_TEXTURE_2D, texnum);
}
void ResourceManager::loadSprites()
{
	// Load the HUD & faces
	if (! gameOptions.disableHud) {
		loadHud();
	}
	spriteSets.resize(SpriteSetMax + 1);
	loadSpriteSet("gfx/objects.tga", SpriteSetMisc, 96);
	loadSpriteSet("gfx/mainchar_pink.tga", SpriteSetMainChar, 68);
	loadSpriteSet("gfx/enemies_pink.tga", SpriteSetEnemies, 68);
	generateSprites();
}

void ResourceManager::generateSprites() {
	BITMAP *temp = create_bitmap_ex(24, 32, 32);
	for (int i=0; i<32; i++) {
		for (int j=0; j<32; j++) {
			float dist = ((i - 16) * (i - 16)) + ((j - 16) * (j - 16));
			dist = sqrt(dist);
			float brightness = 255 * (1 - (dist / 15));
			if (brightness < 0) brightness = 0;
			int colour = makecol24((int) brightness, (int) brightness, (int) brightness);
			putpixel(temp,i,j,colour);
		}
	}
	GLuint newsprite = allegro_gl_make_texture_ex( 0, // flags (AGL_TEXTURE_MIPMAP)
			temp,
			GL_RGB);
	if (newsprite == 0) {
		throw "can't make generated sprite";
	}
	spriteSets[SpriteSetMisc].push_back(newsprite);
	destroy_bitmap(temp);
	
}

void ResourceManager::loadHud()
{
	PALETTE pal;
	BITMAP * hudbmp = load_bitmap("gfx/hud.tga", pal);
	if (! hudbmp) {
		throw "Can't load hud";
	}
	BITMAP * temp = create_bitmap_ex(24, 256,32);
	clear_to_color(temp,0);
	blit(hudbmp, temp, 0,0,0,0, hudbmp->w, hudbmp->h);
	/*
	if (! allegro_gl_check_texture(temp)) {
		throw "Check failed for hud texture";
	}
	*/
	hudtex = allegro_gl_make_texture_ex( 0, // flags (AGL_TEXTURE_MIPMAP)
			temp,
			GL_RGB);
	if (hudtex == 0) {
		throw "failed to make hud texture";
	}
	destroy_bitmap(temp);
	destroy_bitmap(hudbmp);
	
	BITMAP * facesbmp = load_bitmap("gfx/faces.tga", pal);
	if (! facesbmp) {
		throw "Can't load faces";
	}
	temp = create_bitmap_ex(24, 256, 32);
	clear_to_color(temp,0);
	blit(facesbmp, temp, 0,0,0,0, facesbmp->w, facesbmp->h);
	facestex = allegro_gl_make_texture_ex( 0, // flags (AGL_TEXTURE_MIPMAP)
			temp,
			GL_RGB);
	if (facestex == 0) {
		throw "failed to make faces texture";
	}
	destroy_bitmap(temp);
	destroy_bitmap(facesbmp);
}

/* 
 * Because we can't (on some cards) make 96x96 textures, use 128x128 instead and place
 * the sprite in the middle.
 *
 * This wastes vram. Too bad.
 */
void ResourceManager::loadSpriteSet(const char *filename, SpriteSet set, int spritesize)
{
	// Load the bitmpa.
	PALETTE pal;
	BITMAP * bmp = load_bitmap(filename, pal);
	if (! bmp) {
		throw "Can't load sprite set";
	}
	int w,h;
	w = bmp->w / spritesize;
	h = bmp->h / spritesize;
	BITMAP *temp = create_bitmap_ex(24, 128,128);
	int magicpink = makecol24(255,0,255);
	int border = (128 - spritesize) / 2;
	for (int j=0; j<h; j++) {
		for (int i=0; i<w; i++) {
			clear_to_color(temp, magicpink);
			// blit the relevant portion.
			blit(bmp, temp, (i * spritesize), (j * spritesize), border, border, 
				spritesize, spritesize);
			// Now make the opengl texture.
			GLuint texnum = allegro_gl_make_texture_ex( AGL_TEXTURE_MASKED, // flags
				temp,
				GL_RGBA);
			if (texnum == 0) {
				throw "Unable to convert spriteset texture.";
			}
			spriteSets[set].push_back(texnum);
		}
	}
	destroy_bitmap(temp);
	destroy_bitmap(bmp);
}


void ResourceManager::selectSprite(SpriteSet set, int n)
{
	GLuint texnum = spriteSets[set][n];
        glBindTexture(GL_TEXTURE_2D, texnum);
}

